/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

#version 120
uniform vec4 data[32];
uniform int count;
varying vec4 diffuse, ambient;
varying vec3 normal, lightDir, halfVector;

vec4 HSVtoRGB(float h, float s, float v)
{
    int i;
    float f, p, q, t;
    vec4 RGB;

    if (s == 0.0)
    {
        // achromatic (grey)
        RGB.x = RGB.y = RGB.z = v;
        RGB.w = 1.0;
        return RGB;
    }

    h /= 60.0; // sector 0 to 5
    h = min(h, 5.99999);
    i = int(floor(h));

    f = h - float(i); // factorial part of h
    p = v * (1.0 - s);
    q = v * (1.0 - s * f);
    t = v * (1.0 - s * (1.0 - f));

   if(i==0)
    {
        RGB.x = v;
        RGB.y = t;
        RGB.z = p;
    }
    else if(i==1)
    {
        RGB.x = q;
        RGB.y = v;
        RGB.z = p;
    }
    else if(i==2)
    {
        RGB.x = p;
        RGB.y = v;
        RGB.z = t;
    }
    else if(i==3)
    {
        RGB.x = p;
        RGB.y = q;
        RGB.z = v;
    }
    else if(i==4)
    {
        RGB.x = t;
        RGB.y = p;
        RGB.z = v;
    }
    else if(i==5)
    {
        RGB.x = v;
        RGB.y = p;
        RGB.z = q;
    }
    else
    {
        RGB.x = 1.0;
        RGB.y = 1.0;
        RGB.z = 1.0;
    }
    RGB.w = 1.0;
    return RGB;
}

float PI = atan(1.0)*4.0;
//Note for later imporvment:
//The values for smaller m could b cache, since they are already computed
//and proaply will be needed for derivatives or other terms.

float plegendre(const int l, const int m, const float x)
{
    int i, ll;
    float fact, oldfact, pll, pmm, pmmp1, omx2;
    pmm = 1.0;
    if (m > 0)
    {
        omx2 = (1.0 - x) * (1.0 + x);
        fact = 1.0;
        for (i = 1; i <= m; i++)
        {
            pmm *= omx2 * fact / (fact + 1.0);
            fact += 2.0;
        }
    }
    pmm = sqrt(float(2 * m + 1) * pmm / (4.0 * PI));
    if (int(mod(float(m), 2.0)) == 1)
        pmm = -pmm;
    if (l == m)
        return pmm;
    else
    {
        pmmp1 = x * sqrt(float(2 * m + 3)) * pmm;
        if (l == (m + 1))
            return pmmp1;
        else
        {
            oldfact = sqrt(float(2 * m + 3));
            pll = 0.0;
            for (ll = m + 2; ll <= l; ll++)
            {
                fact = sqrt(float(4 * ll * ll - 1) / float(ll * ll - m * m));
                pll = (x * pmmp1 - pmm / oldfact) * fact;
                oldfact = fact;
                pmm = pmmp1;
                pmmp1 = pll;
            }
            return pll;
        }
    }
}


void compute(const int l, const int m, const float cos_theta, const float phi,
        out vec2 val, out vec2 dtheta, out vec2 dphi)
{   
    int m_abs = int(abs(float(m)));
    float p = plegendre(l, m_abs, cos_theta);

    float val_mag = p;
    if (m<0 && int(mod(float(m_abs), 2.0)) == 1)
        val_mag = -val_mag;

    float arg = float(m) * phi;    
    
    float dtheta_mag;
    if (m_abs == 0)
    {
        if (l == 0)
            dtheta_mag = 0.0;
        else
            dtheta_mag = sqrt(float(l * (l + 1))) * plegendre(l, 1, cos_theta);

    }
    else if (m_abs == l)
    {
        dtheta_mag = -sqrt(float(l) / 2.0) * plegendre(l, l - 1, cos_theta);
    }
    else
    {
        dtheta_mag = 0.5 * sqrt(float((l - m_abs) * (l + m_abs + 1)))
                 * plegendre(l, m_abs + 1, cos_theta)
           - 0.5 * sqrt(float((l + m_abs) * (l - m_abs + 1)))
                 * plegendre(l, m_abs - 1, cos_theta);
    }

    if (m<0 && int(mod(float(m_abs), 2.0)) == 1)
        dtheta_mag = -dtheta_mag;
    
    val.x = val_mag * cos(arg);
    val.y = val_mag * sin(arg);
    
    dtheta.x = dtheta_mag * cos(arg);
    dtheta.y = dtheta_mag * sin(arg);
    
    dphi.x = m*val_mag * cos(arg+PI/2.0);
    dphi.y = m*val_mag * sin(arg+PI/2.0);       
}



float cAbs(vec2 z)
{
    return sqrt(z.x * z.x + z.y * z.y);
}

float cArg(vec2 z)
{
    return atan(z.y, z.x);
}

vec2 cMul(vec2 z1, vec2 z2)
{
    return vec2(z1.x * z2.x - z1.y * z2.y, z1.x * z2.y + z1.y * z2.x);
}

/*
We are plotting the following isosurface: F(r, theta, phi) = 0
where F(r, theta, phi) = abs(Y(theta, phi))-r
We color each point (theta, phi) according to: arg(Y(theta, phi))
*/
void main()
{
    float r = length(vec3(gl_Vertex));
    float phi = atan(gl_Vertex.y, gl_Vertex.x);
    float cos_theta = gl_Vertex.z / r;
    float theta = acos(cos_theta);
    float sin_theta = sin(theta);
    float sin_phi = sin(phi);
    float cos_phi = cos(phi);
    
    vec2 total_val = vec2(0);
    vec2 total_dtheta = vec2(0);
    vec2 total_dphi = vec2(0);

    for(int i=0;i<count ;i++)
    {
        int l = int(data[i].x);
        int m = int(data[i].y);
        vec2 c = vec2(data[i].z, data[i].w);
        vec2 val, dtheta, dphi;
        compute(l, m, cos_theta, phi, val, dtheta, dphi);
        total_val = total_val + cMul(c, val);
        total_dtheta = total_dtheta + cMul(c, dtheta);
        total_dphi = total_dphi + cMul(c, dphi);
    }

    float val_mag = cAbs(total_val);
    float val_arg = cArg(total_val);
    
    //Modify vertex
    vec4 newVertex = val_mag * gl_Vertex;
    newVertex.w = 1.0;

    //Compute color
    val_arg = degrees(val_arg);
    while (val_arg < 0.0)
        val_arg += 360.0;
    while (val_arg >= 360.0)
        val_arg -= 360.0;
    vec4 newColor = HSVtoRGB(val_arg, 1.0, 1.0);

    //Compute normal
    vec3 rv = vec3(gl_Vertex) / r;
    vec3 thetav = vec3(cos_theta * cos_phi, cos_theta * sin_phi, -sin_theta);
    vec3 phiv = vec3(-sin_phi, cos_phi, 0);
    
    vec2 val_conj = vec2(total_val.x, -1.0*total_val.y);
    vec2 temp;    
    temp = cMul(total_dtheta, val_conj);
    float modulus_dtheta = temp.x/val_mag;
    temp = cMul(total_dphi, val_conj);    
    float modulus_dphi = temp.x/val_mag;
    
    vec3 newNormal = rv - (modulus_dtheta/r) * thetav  
                        - (modulus_dphi/(r*sin_theta)) * phiv;


    /* first transform the normal into eye space and normalize the result */
    normal = normalize(gl_NormalMatrix * newNormal);

    /* now normalize the light's direction. Note that according to the
     OpenGL specification, the light is stored in eye space. Also since 
     we're talking about a directional light, the position field is actually 
     direction */

    lightDir = normalize(vec3(gl_LightSource[0].position));

    /* Normalize the halfVector to pass it to the fragment shader */
    halfVector = normalize(gl_LightSource[0].halfVector.xyz);

    /* Compute the diffuse, ambient and globalAmbient terms */
    diffuse = newColor * gl_LightSource[0].diffuse;
    ambient = gl_FrontMaterial.ambient * gl_LightSource[0].ambient;
    ambient += gl_LightModel.ambient * gl_FrontMaterial.ambient;

    gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * newVertex;

}
